import pandas
import feather


def predict_model(product, task, params, target_labels=['predictions']):
    if params['split_type'] == 'train_test':
        with open(task.upstream['split_data'].product['test_y_data'], 'rb') as f:
            target_data = feather.read_dataframe(f)
    elif params['split_type'] == 'k_fold':
        target_data = []
        for i in range(1, params['n_folds']+1):
            with open(task.upstream['split_data'].product[f'fold_{i}_y_data'], 'rb') as f:
                target_data.append(
                    feather.read_dataframe(f)
                )
        target_data = pandas.concat(target_data)

    target_shape = target_data.shape
    predictions = feather.read_dataframe(product['predictions'])

    if 'quantile_intervals' in params:
        quantiles = set()
        for interval in params['quantile_intervals']:
            quantiles.update(interval)
        for quantile in quantiles:
            str_quantile = 'Q' + f'{quantile * 100:g}'
            assert str_quantile in predictions.columns, f'{str_quantile} missing from predictions. Columns are {predictions.columns}.'

    try:
        assert target_shape == predictions[target_labels].shape, f'The predictions should have the same shape as the test data. Currently {predictions.shape} and {target_shape}.'
    except KeyError:
        raise Exception(f'{target_labels} are not in the columns. Columns are {predictions.columns}.')



import mvf.integration.config as on_render
import mvf.integration.process as on_finish
import mvf.process as process
import os
from pathlib import Path
from ploomber import DAG
from ploomber.products import File
from ploomber.tasks import PythonCallable, NotebookRunner


class DagBuilder:
    def __init__(self, config, output_dir='output') -> None:
        '''
        Assigns key parameters as attributes. Initialises the Ploomber DAG.
        '''
        self.config = config
        self.output_dir = output_dir
        # path to source code
        self.path_to_process = process.__path__[0]
        self.dag = DAG(
            # set dag name as basename of working dir
            name=os.path.basename(os.getcwd())
        )


    def build(self):
        '''
        Main method for the class.

        Builds ploomber DAG from config file.
        '''
        # build up generic tasks
        preprocess_data = self.__build_preprocess_data()
        split_data = self.__build_split_data()
        validate = self.__build_validate()
        # model tasks
        for model in self.config['models']:
            # get model params
            name = model['name']
            lang = model['lang']
            val_step = model['validation_step']
            return_quantiles = model.get('return_quantiles', False)
            # build up model tasks
            fit_model = self.__build_fit_model(name, lang)
            predict_model = self.__build_predict_model(name, lang, return_quantiles=return_quantiles)
            if val_step:
                validate_model = self.__build_validate_model(name, lang)
                # set upstream
                fit_model >> validate_model
            # set upstream
            split_data >> fit_model
            split_data >> predict_model
            fit_model >> predict_model
            fit_model >> validate
            predict_model >> validate
        # set upstream
        preprocess_data >> split_data
        split_data >> validate


    def __build_preprocess_data(self):
        # define task
        preprocess_data = NotebookRunner(
            source=Path(
                os.path.abspath(
                    self.config['data']['source']
                )
            ),
            product={
                'nb': File(
                    os.path.join(
                        self.output_dir,
                        'notebooks',
                        self.config['data']['source']
                    ),
                ),
                'X_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        'preprocess_X_data.feather'
                    ),
                ),
                'y_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        'preprocess_y_data.feather'
                    ),
                ),
            },
            dag=self.dag,
            name='preprocess_data',
        )
        # hooks
        preprocess_data.on_render = on_render.preprocess_data.preprocess_data
        preprocess_data.on_finish = on_finish.preprocess_data.preprocess_data
        return preprocess_data


    def __build_split_data(self):
        # params
        params = {
            'split_type': self.config['data']['split']
        }
        # define params based on split_type
        if self.config['data']['split'] == 'train_test':
            params['test_size'] = self.config['data']['test_size']
            product = {
                'train_X_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        'train_X_data.feather'
                    ),
                ),
                'test_X_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        'test_X_data.feather'
                    ),
                ),
                'train_y_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        'train_y_data.feather'
                    ),
                ),
                'test_y_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        'test_y_data.feather'
                    ),
                ),
            }
        else:
            n_folds = self.config['data']['n_folds']
            params['n_folds'] = n_folds
            product = {
                f'fold_{i}_X_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        f'fold_{i}_X_data.feather'
                    ),
                )
                for i in range(1, n_folds + 1)
            }
            product_y = {
                f'fold_{i}_y_data': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        f'fold_{i}_y_data.feather'
                    ),
                )
                for i in range(1, n_folds + 1)
            }
            product.update(product_y)
        # define task
        split_data = PythonCallable(
            source=process.split_data.split_data,
            product=product,
            dag=self.dag,
            name='split_data',
            params=params,
        )
        # hooks
        split_data.on_render = on_render.split_data.split_data
        split_data.on_finish = on_finish.split_data.split_data
        return split_data
    

    def __build_fit_model(self, name, lang):
        task_name = name + '_fit'
        # params
        params = {
            'model_name': name,
            'split_type': self.config['data']['split'],
            'lang': lang,
        }
        if self.config['data']['split'] == 'k_fold':
            params['split_type'] = self.config['data']['split']
            params['n_folds'] = self.config['data']['n_folds']
            product = {
                f'model_{i}': File(
                    os.path.join(
                        self.output_dir,
                        'models',
                        task_name + f'_{i}'
                    ),
                )
                for i in range(1, self.config['data']['n_folds'] + 1)
            }
        else:
            product = {
                'model': File(
                    os.path.join(
                        self.output_dir,
                        'models',
                        task_name
                    ),
                ),
            }
        product['process_metadata'] = File(
            os.path.join(
                self.output_dir,
                'metadata',
                f'{task_name}.metadata'
            )
        )
        # define task
        fit_model = PythonCallable(
            source=process.fit_model.fit_model,
            product=product,
            dag=self.dag,
            name=task_name,
            params=params,
        )
        # hooks
        if lang == 'Python':
            fit_model.on_render = on_render.fit_model.fit_model_py
            fit_model.on_finish = on_finish.fit_model.fit_model_py
        else:
            fit_model.on_render = on_render.fit_model.fit_model_r
            fit_model.on_finish = on_finish.fit_model.fit_model_r
        return fit_model


    def __build_predict_model(self, name, lang, return_quantiles=False):
        task_name = name + '_predict'
        # source
        source = process.predict_model.predict_model
        # params
        params = {
            'split_type': self.config['data']['split'],
            'model_name': name,
            'lang': lang,
        }
        if self.config['data']['split'] == 'k_fold':
            params['n_folds'] = self.config['data']['n_folds']
        if 'output' in self.config:
            if 'quantile_intervals' in self.config['output'] and return_quantiles:
                params['quantile_intervals'] = self.config['output']['quantile_intervals']
        # define task
        predict_model = PythonCallable(
            source=source,
            product={
                'predictions': File(
                    os.path.join(
                        self.output_dir,
                        'data',
                        f'{task_name}.feather'
                    ),
                ),
                'process_metadata': File(
                    os.path.join(
                        self.output_dir,
                        'metadata',
                        f'{task_name}.metadata'
                    )
                )
            },
            dag=self.dag,
            name=task_name,
            params=params
        )
        # hooks
        if lang == 'Python':
            predict_model.on_render = on_render.predict_model.predict_model_py
        else:
            predict_model.on_render = on_render.predict_model.predict_model_r
        predict_model.on_finish = on_finish.predict_model.predict_model
        return predict_model
    

    def __build_validate_model(self, name, lang):
        task_name = name + '_validate'
        # params
        params = {
            'model_name': name,
            'split_type': self.config['data']['split'],
            'lang': lang
        }
        if self.config['data']['split'] == 'k_fold':
            params['n_folds'] = self.config['data']['n_folds']
        # define task
        validate_model = NotebookRunner(
            source=Path(
                os.path.join(
                    self.path_to_process,
                    'validate_model.py'
                )
            ),
            product={
                'nb': File(
                    os.path.join(
                        self.output_dir,
                        'notebooks',
                        task_name + '.html'
                    ),
                ),
            },
            dag=self.dag,
            name=task_name,
            params=params,
        )
        # hooks
        if lang == 'Python':
            validate_model.on_render = on_render.validate_model.validate_model_py
        else:
            validate_model.on_render = on_render.validate_model.validate_model_r
        validate_model.on_finish = on_finish.validate_model.validate_model
        return validate_model


    def __build_validate(self):
        source_path = Path(
            os.path.join(
                self.path_to_process,
                'validate.py'
            )
        )
        # params
        params = {
            'split_type': self.config['data']['split'],
            'models': [model['name'] for model in self.config['models']]
        }
        if self.config['data']['split'] == 'k_fold':
            params['n_folds'] = self.config['data']['n_folds']
        if 'output' in self.config:
            if 'quantile_intervals' in self.config['output']:
                params['quantile_intervals'] = self.config['output']['quantile_intervals']
        # define task
        validate = NotebookRunner(
            source=source_path,
            product={
                'nb': File(
                    os.path.join(
                        self.output_dir,
                        'notebooks',
                        'validate.ipynb'
                    ),
                ),
                'nb_html': File(
                    os.path.join(
                        self.output_dir,
                        'notebooks',
                        'validate.html'
                    ),
                ),
            },
            nb_product_key=[
                'nb',
                'nb_html'
            ],
            dag=self.dag,
            name='validate',
            params=params,
        )
        # hooks
        validate.on_render = on_render.validate.validate
        validate.on_finish = on_finish.validate.validate
        return validate
    


def fit_model_params(product: dict, params: dict) -> None:
    # check params for fit_model tasks
    assert 'model_name' in params, 'The \'model_name\' parameter must be defined.'
    assert 'split_type' in params, 'The \'split_type\' parameter must be defined.'
    if params['split_type'] == 'train_test':
        assert 'model' in product, 'The \'model\' product must be defined.'
    elif params['split_type'] == 'k_fold':
        assert 'n_folds' in params, 'For a K fold split, the n_folds must be defined.'
        n_folds = params['n_folds']
        for i in range(1, n_folds+1):
            assert f'model_{i}' in product, f'The {i}-th model product must be defined.'


def fit_model_r(product: dict, params: dict) -> None:
    import rpy2.robjects as robjects
    import rpy2_r6.r6b as r6b

    # check params
    fit_model_params(product, params)
    # check that the correct model class is available
    r = robjects.r
    try:
        r.source('models.R')
        model_class = r6b.R6DynamicClassGenerator(r[params["model_name"]])
        model = model_class.new()
        check_model(model, params['model_name'])
    except rpy2.rinterface_lib.embedded.RRuntimeError:
        raise ModuleNotFoundError(f'If using a R based model for {params["model_name"]}, you must have a models.R file in your project directory.')


def fit_model_py(product: dict, params: dict) -> None:
    # check params
    fit_model_params(product, params)
    # check that the correct model class is available
    try:
        import models
        assert hasattr(models, params['model_name']), f'models.py must contain a class called {params["model_name"]}.'
        model_class = getattr(models, params['model_name'])
        check_model(model_class, params['model_name'])
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f'If using a Python based model for {params["model_name"]}, you must have a models.py file in your project directory. The project directory is {os.getcwd()}')
    

def check_model(model_class, model_name):
    '''
    Checks the instantiated model class or generic model class has the necessary methods.
    '''
    # check the model class has a callable 'fit' method
    assert hasattr(model_class, 'fit'), f'The class {model_name} must have a fit() method. The class does not currently have a fit attribute.'
    assert callable(getattr(model_class, 'fit')), f'The class {model_name} must have a fit() method. The class\' fit attribute is not currently callable.'
    # check the model class has a callable 'save' method
    assert hasattr(model_class, 'save'), f'The class {model_name} must have a save() method. The class does not currently have a save attribute.'
    assert callable(getattr(model_class, 'save')), f'The class {model_name} must have a save() method. The class\' save attribute is not currently callable.'
    


import numbers

def split_data(product: dict, params: dict) -> None:
    # check split_type
    assert 'split_type' in params, 'The \'split_type\' parameters must be provided.'
    supported_split_types = ['train_test', 'k_fold']
    assert params['split_type'] in supported_split_types, f"{params['split_type']} is not supported. Supported split types are {supported_split_types}."
    # given split, type check accompanying parameters
    if params['split_type'] == 'train_test':
        # check injection params
        assert 'test_size' in params, 'If the split_type is \'train_test\', the test_size parameter must be provided.'
        assert isinstance(params['test_size'], numbers.Number), f"test_size must be a number. Currently {type(params['test_size'])}."
        assert 0 <= params['test_size'] <= 1, f"test_size must be between 0 and 1. Currently {params['test_size']}."
        # check product params
        assert 'train_X_data' in product, 'The \'train_X_data\' product must be defined.'
        assert 'test_X_data' in product, 'The \'test_X_data\' product must be defined.'
        assert 'train_y_data' in product, 'The \'train_y_data\' product must be defined.'
        assert 'test_y_data' in product, 'The \'test_y_data\' product must be defined.'
    # else must be 'k_fold'
    else:
        # check injection params
        assert 'n_folds' in params, 'If the split_type is \'k_fold\', the n_folds parameter must be provided.'
        assert isinstance(params['n_folds'], int), f"n_folds must be a number. Currently {type(params['n_folds'])}."
        assert params['n_folds'] >= 2, f"n_folds must be at least 2. Currently {params['n_folds']}."
        # check product params
        for i in range(1, params['n_folds'] + 1):
            assert f'fold_{i}_X_data' in product, f'The \'fold_{i}_X_data\' product must be defined.'
            assert f'fold_{i}_y_data' in product, f'The \'fold_{i}_y_data\' product must be defined.'


import click
import os
import sys
from mvf.cli import utils
from mvf.dag.builder import DagBuilder
from mvf.template.config import ConfigBuilder
from rpy2.rinterface_lib.callbacks import logger as rpy2_logger
import logging
rpy2_logger.setLevel(logging.ERROR)


@click.group(
    help='''
    MVF

    See the documentation: 

        https://tomkimcta.gitlab.io/model-validation-framework
    '''
)
def mvf():
    # '''
    # CLI entry point.
    # '''
    # add working dir to PYTHONPATH to allow import of local modules
    sys.path.append(os.getcwd())


@click.command(
    help='''
        Utility for initialising an MVF project. Generates an MVF
        configuration file. Default use will initialise the MVF
        project in the current working directory.
    '''
)
@click.option(
    '--directory',
    '-d',
    default='.',
    help='The directory in which to initialise the MVF project.'
)
def init(directory):
    # check if default use
    if directory == '.':
        pth = os.getcwd()
        click.echo('Initialising MVF project in the current working directory')
    # path specified by user
    else:
        if os.path.isabs(directory):
            pth = directory
        # if relative path provided, construct absolute
        else:
            pth = os.path.join(
                os.getcwd(),
                directory
            )
        # check if the path provided by the user is already a file
        if os.path.isfile(pth):
            raise click.ClickException('A file already exists at the location provided!')
        click.echo(f'Initialising MVF project in {pth}')
    # get confirmation from user
    click.confirm(f'Do you want to continue?', abort=True)
    cfg_builder = ConfigBuilder(pth)
    cfg_builder.write()


@click.command(
    help='''
        Utility for running MVF projects or processes. Default use
        will trigger an incremental execution of the project.
    '''
)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help="Force execution of whole workflow."
)
@click.option(
    '--output',
    '-o',
    default='output',
    help='Specify the output directory.'
)
@click.option(
    '--process',
    '-p',
    default=None,
    help='Force execution of a specific process by name.'
)
def run(force, output, process):
    # load project config
    config = utils.load_config(
        os.path.join(
            os.getcwd(), 
            'mvf_conf.yaml'
        )
    )
    # build dag from config
    click.echo('Building MVF project workflow...')
    dag_builder = DagBuilder(config, output_dir=output)
    dag_builder.build()
    if process is None:
        # access the built dag and execute it
        click.echo('Running MVF project...')
        dag_builder.dag.build(force=force)
    else:
        # access a process and execute it
        dag_process = dag_builder.dag.get(process)
        dag_builder.dag.render(force=force)
        click.echo(f'Running {process} process...')
        dag_process.build(force=True)
    # plot the dag
    click.echo('Plotting workflow...')
    dag_builder.dag.plot('project_workflow.html')


@click.command(
    help='''
        Utility for plotting MVF workflows. If saving outputs to a
        custom directory, this directory should be passed as an
        option to get up-to-date process statuses.
    '''
)
@click.option(
    '--output',
    '-o',
    default='output',
    help='Specify the output directory.'
)
def plot(output):
    # load project config
    config = utils.load_config(
        os.path.join(
            os.getcwd(), 
            'mvf_conf.yaml'
        )
    )
    # build dag from config
    dag_builder = DagBuilder(config, output_dir=output)
    dag_builder.build()
    click.echo('Plotting workflow...')
    # access the built dag and plot it
    dag_builder.dag.plot('project_workflow.html')


# add commands to group
mvf.add_command(run)
mvf.add_command(plot)
mvf.add_command(init)



import feather

def split_data(product: dict, params: dict):
    # different tests by split type
    if params['split_type'] == 'train_test':
        X_train = feather.read_dataframe(product['train_X_data'])
        X_test = feather.read_dataframe(product['test_X_data'])
        y_train = feather.read_dataframe(product['train_y_data'])
        y_test = feather.read_dataframe(product['test_y_data'])

        assert X_train.shape[0] == y_train.shape[0], f'X and y train data should have the same number of instances. Currently {X_train.shape[0]} and {y_train.shape[0]}'
        assert X_test.shape[0] == y_test.shape[0], f'X and y test data should have the same number of instances. Currently {X_test.shape[0]} and {y_test.shape[0]}'
        assert X_train.shape[1] == X_test.shape[1], f'X train and test data should have the same number of features. Currently {X_train.shape[1]} and {X_test.shape[1]}'
        assert y_train.shape[1] == y_test.shape[1], f'y train and test data should have the same number of features. Currently {y_train.shape[1]} and {y_test.shape[1]}'
        assert (X_test.shape[0] / (X_test.shape[0] + X_train.shape[0])) - params['test_size'] < 1e-4, 'The actual test size differs too much from the test_size parameter. If you have a very small dataset, choose an appropriate test size for the number of data points.'
        assert (X_train.index == y_train.index).all(), 'The indices for X and y training data do not match up.'
        assert (X_test.index == y_test.index).all(), 'The indices for X and y testing data do not match up.'
        assert len(set(X_train.index).intersection(set(X_test.index))) == 0, 'The train and test indices are not distinct.'
    elif params['split_type'] == 'k_fold':
        X_data = []
        y_data = []
        n_instances = set()
        indices = []
        for i in range(1, params['n_folds'] + 1):
            fold_i_X_data = feather.read_dataframe(product[f'fold_{i}_X_data'])
            fold_i_y_data = feather.read_dataframe(product[f'fold_{i}_y_data'])

            assert fold_i_X_data.shape[0] == fold_i_y_data.shape[0], f'X and y data from a particular fold should have the same number of instances. Currently {fold_i_X_data.shape[0]} and {fold_i_y_data.shape[0]} for the {i}th fold.'
            assert (fold_i_X_data.index == fold_i_y_data.index).all(), f'The indices for X and y from the {i}th fold do not match up.'
            
            X_data.append(fold_i_X_data)
            y_data.append(fold_i_y_data)
            n_instances.add(fold_i_X_data.shape[0])
            indices.append(set(fold_i_X_data.index))

        assert len(n_instances) <= 2 and -1 <= list(n_instances)[0] - list(n_instances)[-1] <= 1, f'The number of instances in each fold must be as equal as possible. The number of instances in folds are currently {n_instances}.'
        assert len(set.intersection(*indices)) == 0, 'The fold indices are not distinct.'

        for i in range(params['n_folds'] - 1):
            assert X_data[i].shape[1] == X_data[i+1].shape[1], f'X data should have the same number of features across all folds. {i}th fold has {X_data[i].shape[1]} features but {i+1}th fold has {X_data[i+1].shape[1]} features.'
            assert y_data[i].shape[1] == y_data[i+1].shape[1], f'y data should have the same number of features across all folds. {i}th fold has {y_data[i].shape[1]} features but {i+1}th fold has {y_data[i+1].shape[1]} features.'
    else:
        raise NotImplementedError(f"The {params['split_type']} implementation is not tested. This is unacceptable.")



# imports
from mvf.process.utils import import_model_class, init_model

# + tags=["parameters"]
upstream = None
product = None
model_name = ''
split_type = 'train_test'
n_folds = 10
lang = ''
# -

# import model module
model_module = import_model_class(lang)

# run validation
if split_type == 'train_test':
    # load model
    model = init_model(
        model_module, 
        lang,
        model_name,
    )
    model.load(str(next(iter(upstream.values()))['model']))
    # run model specific validation
    model.validate()
elif split_type == 'k_fold':
    for i in range(1, n_folds+1):
        # load model
        model = init_model(
            model_module, 
            lang,
            model_name,
        )
        model.load(str(next(iter(upstream.values()))[f'model_{i}']))
        # run model specific validation
        model.validate()


def preprocess_data(product: dict) -> None:
    # check product params
    assert 'X_data' in product, 'The \'X_data\' product must be defined.'
    assert 'y_data' in product, 'The \'y_data\' product must be defined.'


import feather

def preprocess_data(product: dict):
    X = feather.read_dataframe(product['X_data'])
    y = feather.read_dataframe(product['y_data'])
    assert len(X.shape) == 2, f'Expected the X data to have 2 dimensions, has {len(X.shape)}'
    assert len(y.shape) == 2, f'Expected the y data to have 2 dimensions, has {len(y.shape)}'  



import click
import yaml
from mvf.integration import mvf_config

def load_config(path):
    # open the mvf config file
    try:
        with open(path, 'r') as f:
            config = yaml.safe_load(f)
    except FileNotFoundError:
        raise Exception('No `mvf_conf.yaml` found in the working directory.')
    else:
        # validate the config against the schema
        click.echo('Validating config...')
        mvf_config.check_config(config)
        return config


def predict_model_py(product, params):
    import models
    # get model class
    model_class = getattr(models, params['model_name'])
    # init model
    model = model_class()
    # run common checks
    predict_model_common(product, params, model)
    

def predict_model_r(product, params):
    import rpy2.robjects as robjects
    import rpy2_r6.r6b as r6b
    # get model class
    r = robjects.r
    r.source('models.R')
    model_class = r6b.R6DynamicClassGenerator(r[params['model_name']])
    # init model
    model = model_class.new()
    # run common checks
    predict_model_common(product, params, model)
    

def predict_model_common(product, params, model):
    '''
    Method to run checks common to Python and R
    '''
    assert 'predictions' in product, 'The \'predictions\' product must be defined.'
    # check the model has a predict method
    assert hasattr(model, 'predict'), f'The model class defined for {params["model_name"]} must have a predict() method.'
    assert callable(getattr(model, 'predict')), f'The model saved at {params["model_name"]} must have a predict() method. The class\' predict attribute is not currently callable.'
    # check the model has a load method
    assert hasattr(model, 'load'), f'The model class defined for {params["model_name"]} must have a load() method.'
    assert callable(getattr(model, 'load')), f'The model saved at {params["model_name"]} must have a load() method. The class\' load attribute is not currently callable.'



# imports
import feather
from sklearn.model_selection import train_test_split, KFold


def split_data(upstream, product, split_type, test_size=0.5, n_folds=10):
    # load data from upstream process
    X = feather.read_dataframe(upstream['preprocess_data']['X_data'])
    y = feather.read_dataframe(upstream['preprocess_data']['y_data'])

    # parse split_type parameter
    if split_type == 'train_test':
        # split data
        X_train, X_test, y_train, y_test = train_test_split(
            X, 
            y, 
            test_size=test_size
        )
        # print metadata
        print(f'X_train shape: {X_train.shape}')
        print(f'X_test shape: {X_test.shape}')
        print(f'y_train shape: {y_train.shape}')
        print(f'y_test shape: {y_test.shape}')
        # save data for next process
        feather.write_dataframe(X_train, product['train_X_data'])
        feather.write_dataframe(X_test, product['test_X_data'])
        feather.write_dataframe(y_train, product['train_y_data'])
        feather.write_dataframe(y_test, product['test_y_data'])
    elif split_type == 'k_fold':
        kfolds = KFold(n_splits=n_folds)
        for i, (train_idx, test_idx) in enumerate(kfolds.split(X)):
            # split data
            fold_X_data = X.iloc[test_idx]
            fold_y_data = y.iloc[test_idx]
            # print metadata
            print(f'Fold {i+1} X shape: {fold_X_data.shape}')        
            print(f'Fold {i+1} y shape: {fold_y_data.shape}')
            # save data for next process   
            feather.write_dataframe(fold_X_data, product[f'fold_{i+1}_X_data'])   
            feather.write_dataframe(fold_y_data, product[f'fold_{i+1}_y_data'])   



from setuptools import setup, find_packages


def _get_version():
    version = '1.0.0'
    with open('version') as file:
        version = file.read().strip()
    return version


def _get_long_description():
    with open('README.md', 'r') as f:
        long_description = f.read()
    return long_description


setup(
    name='mvf',
    version=_get_version(),
    author='Tom Kim',
    author_email='tom.kim@certus-tech.com',
    description='A package implementing a supervised learning model validation framework.',
    long_description=_get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(
        exclude=[
            'test*',
            'documentation*',
            'examples*',
        ],
    ),
    install_requires=[
        'click',
        'feather-format',
        'pandas',
        'ploomber',
        'rpy2',
        'rpy2-r6',
        'schema',
        'scikit-learn',
    ],
    extras_require={
        'dev': [
            'coverage',
            'pytest',
            'testbook',
            'twine',
        ]
    },
    keywords=[
        'python',
        'R',
        'machine learning',
        'validation',
        'framework',
    ],
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'Framework :: Jupyter',
        'License :: Free For Educational Use',
        'Operating System :: Unix',
    ],
    python_requires='>=3.9',
    entry_points={
        'console_scripts': [
            'mvf=mvf.cli.cli:mvf'
        ]
    },
)



import timeit
import pickle


class ProcessTimer:
    def __init__(self, process_name):
        self.process_name = process_name
        self.start = timeit.default_timer()

    def end(self):
        self.end = timeit.default_timer()

    def save(self, path):
        process_time = {
            self.process_name: self.end - self.start
        }
        with open(path, 'wb') as f:
            pickle.dump(process_time, f, protocol=pickle.HIGHEST_PROTOCOL)


def import_model_class(lang):
    '''
    Imports model module depending on language parameter.
    '''
    if lang == 'Python':
        import models
        return models
    elif lang == 'R':
        import rpy2.robjects as robjects
        from rpy2.robjects import pandas2ri
        import rpy2_r6.r6b as r6b

        r = robjects.r
        r['source']('models.R')

        return {
            'r6b': r6b,
            'r': r,
            'pandas2ri': pandas2ri,
            'robjects': robjects
        }


def init_model(model_module, lang, model_name):
    '''
    Instantiates a model.
    '''
    if lang == 'Python':
        model_class = getattr(model_module, model_name)
        model = model_class()
    elif lang == 'R':
        r6b = model_module['r6b']
        r = model_module['r']
        model_class = r6b.R6DynamicClassGenerator(r[model_name])
        model = model_class.new()
    return model
